/**
 * ----------------------------------------------------------------------------
 * Programmation d'étiquettes RFID / RC522 a l'aide d'un module I2C
 * Copyright Locoduino et Dominique Bultez 2020
 * Version 1.1 du 4 juilet 2020
 *
 * Attention: des données vont être écrites dans le secteur #1 (blocks #4 à #7).
 * Format :
 * NN BB63259xxxxxx
 * NN numéro de loco ou train (l'étiquette n'est pas forcément sous la loco),
 * puis un espace séparateur,
 * puis le nom du train ou loco,
 * La saisie se termine avec un "#" puis valider.
 * 
 * Le caractère # si présent, permet l'affichage immédiat de la saisie
 * mais, s'il est oublié, elle se produit quand même au bout de 10 secondes environ après validation.
 *
 * Seules les étiquettes Mifare Ultralight sont supportées
 * 
 * Après programmation d'une étiquette avec succès, il faut appuyer sur le bouton RESET
 * pour programmer une autre étiquette.
 */
 
#include <Wire.h>
#include "MFRC522_I2C.h"
//https://github.com/arozcan/MFRC522-I2C-Library

const int RST_PIN = 6;    // Arduino Nano Pin ou autre Pin au choix

// 0x28 est l'addresse i2c . Chercher avec i2c_scanner si inconnue.

MFRC522 mfrc522(0x28, RST_PIN);   // Creation d'une instance MFRC522 
MFRC522::StatusCode status;       // variable pour le statut des opérations 

  byte dataBlock[]    = {       //  saisie des données à enregistrer sur l'étiquette
        0x00, 0x00, 0x00, 0x00, //  0,0,0,0
        0x00, 0x00, 0x00, 0x00, //  0,0,0,0
        0x00, 0x00, 0x00, 0x00, //  0,0,0,0
        0x00, 0x00, 0x00, 0x00  //  0,0,0,0
    };

  uint8_t pageAddr = 0x04;  //Les données sont ecrites dans 4 pages : 4,5,6 et 7.
                            //Pages 0 to 3 sont réservées donc inutilisées.           


/**
 * Routines d'affichage d'un tableau d'octets en Hexa et en Ascii (char)
 */
void dump_byte_array(byte *buffer, byte bufferSize) {
    for (byte i = 0; i < bufferSize; i++) {
        Serial.print(buffer[i] < 0x10 ? " 0" : " ");
        Serial.print(buffer[i], HEX);
    }
}

void dump_char_array(byte *buffer, byte bufferSize) {
    for (byte i = 0; i < bufferSize; i++) {
        Serial.print("  ");
        Serial.print((char)buffer[i]);
    }
}

/**
 * Initialisation.
 */
void setup() {
  Serial.begin(115200);     // Intilatise la connexion serie avec le moniteur
  Wire.begin();             // Initialise I2C
  mfrc522.PCD_Init();       // Initialise le module MFRC522
  Serial.println();
  Serial.println(F("Programmation d'une étiquette MIFARE Ultralight."));
  Serial.println(F("Attention: Ecriture de données dans le secteur #1 (pages 4 à 7)"));
  boolean done = false;
  while (!done) {
    Serial.println(F("entrer le numero de loco (2 car) puis espace puis Nom (7 car maxi) puis '#' puis valider"));
    byte len;
    Serial.setTimeout(10000L) ;       // temps de saisie limitée à 10s si on oublie le #
    while (Serial.available() == 0);  // attente entrée clavier et validation quelque soit le choix de fin de ligne
    len = Serial.readBytesUntil('#', (char *) dataBlock, 16) ; // lecture numero de loco et nom, maxi 16.
    Serial.print(F("Vous avez entré le Numero de loco et le Nom suivant: "));
    Serial.println();
    dump_byte_array(dataBlock, 16); // version Hexa
    Serial.println();
    dump_char_array(dataBlock, 16); // version Ascii
    Serial.println();
    Serial.println(F("taper Y pour programmer une etiquette ou N pour recommencer."));
    Serial.println("---------");
    byte incomingByte;
    while(1) {  //permet de quitter le setup et d'aller dans la loop qui attend l'étiquette et la programme
      if (Serial.available() > 0) {
        // Attente un caractère:
        incomingByte = Serial.read();
        if ((incomingByte == 'Y')||(incomingByte == 'y')) {
          done = true;
          break; // sortie du while
        }
        if ((incomingByte == 'N')|| (incomingByte == 'n')) {
          Serial.println();
          break;  // sortie du while et retour à la saisie
        }
      }
    }
  }
  Serial.println(F("Présentez maintenant une etiquette NTAG213 pour la programmer."));
  Serial.println();
}
/**
 *  Programmation de l'etiquette
 */
void loop() {
    // attente présentation d'une étiquette
    if ( ! mfrc522.PICC_IsNewCardPresent())
        return;
        
    // Selection de l'etiquette
    if ( ! mfrc522.PICC_ReadCardSerial())
        return;
        
    // Affichage de l'UID et du type d'etiquette/tag
    Serial.print(F("Tag UID:"));
    dump_byte_array(mfrc522.uid.uidByte, mfrc522.uid.size);
    Serial.println();
    Serial.print(F("PICC type: "));
    MFRC522::PICC_Type piccType = mfrc522.PICC_GetType(mfrc522.uid.sak);
    Serial.println(mfrc522.PICC_GetTypeName(piccType));
    // verification de compatibilité
    if (piccType != MFRC522::PICC_TYPE_MIFARE_UL) {
        Serial.println(F("Ce programme ne fonctionne qu'avec des étiquettes MIFARE Ultralight"));
        return;
    }
    
    // Nous utilisons les pages 4 à 7 inclus (4 octets/page donc 16 octets)
    byte sector         = 1;
    byte blockAddr      = pageAddr;
    byte trailerBlock   = pageAddr + 3;
    byte buffer[18];        // 16 caractères + CRC
    byte size = sizeof(buffer);
    
     // Affichage du contenu du secteur 
    Serial.print(F("Contenu initial des pages: "));Serial.print(blockAddr);Serial.print(" à ");Serial.println(trailerBlock);
    status = (MFRC522::StatusCode) mfrc522.MIFARE_Read(blockAddr, buffer, &size);
    if (status != MFRC522::STATUS_OK) {
        Serial.print(F("MIFARE_Read() Echec: "));
        Serial.println(mfrc522.GetStatusCodeName(status));
    }
    dump_byte_array(buffer, 16); Serial.println();
    dump_char_array(buffer, 16); Serial.println();
    Serial.println();
    
    // Ecriture des données
    Serial.println(F("Données à écrire :  ")); //Serial.print(blockAddr);
    memcpy(buffer,dataBlock,16);

    //Serial.println(F(" ..."));
    dump_byte_array(buffer, 16); Serial.println();
    dump_char_array(buffer, 16); Serial.println();
    
    Serial.println(F("Ecriture des données..."));
    if (piccType == MFRC522::PICC_TYPE_MIFARE_UL) {
      for (int i=0; i < 4; i++) {
        //les données sont écrites dans une page (4 bytes par page)
        status = (MFRC522::StatusCode) mfrc522.MIFARE_Ultralight_Write(blockAddr+i, &buffer[i*4], 4);
        if (status != MFRC522::STATUS_OK) {
          Serial.print(F("MIFARE_Write() Echec: "));
          Serial.println(mfrc522.GetStatusCodeName(status));
        }
      }
    } 
    Serial.println();
    
    // Lecture et vérification du secteur 
    Serial.println(F("Relecture des données pour vérification... ")); 
    status = (MFRC522::StatusCode) mfrc522.MIFARE_Read(blockAddr, buffer, &size); // &size
    if (status != MFRC522::STATUS_OK) {
        Serial.print(F("MIFARE_Read() Echec: "));
        Serial.println(mfrc522.GetStatusCodeName(status));
    }
    Serial.println(F("Données lues dans l'etiquette; "));
    dump_byte_array(buffer, 16); Serial.println();
    dump_char_array(buffer, 16); Serial.println();
    // On va compter le nombre d'octets identiques à ceux de l'original
    Serial.println(F("Verification..."));
    byte count = 0;
    for (byte i = 0; i < 16; i++) {
        // Compare buffer (= ce qu'on a lu) avec dataBlock (= ce qu'on a écrit)
        if (buffer[i] == dataBlock[i])
            count++;
    }
    Serial.print(F("Nombre d'octets corrects = ")); Serial.println(count);
    if (count == 16) {
        Serial.println(F("Succes :-)"));
    } else {
        Serial.println(F("Echec :-("));
    }
    Serial.println();
    
    // Halt PICC
    mfrc522.PICC_HaltA();
    // Stop encryption on PCD
    mfrc522.PCD_StopCrypto1();
    Serial.println(F("taper Y pour refaire la programmation ou appuyer sur le bouton de RESET pour programmer une autre étiquette"));
    Serial.println();
    byte incomingByte;
    while(1) {  
      if (Serial.available() > 0) {
    // read the incoming byte:
    incomingByte = Serial.read();
    if ((incomingByte == 'Y')||(incomingByte == 'y')) break;
    }
  }
}
